using System;
using System.Collections.Generic;
using  DataContracts;

namespace Ato.EN.IntegrationServices.CodeGenerationCTR
{
    public partial class CTR2022Validator
    {

        #region Manual Rules - Rules marked for manual coding

        #region VR.ATO.CTR.500203

        /*  VR.ATO.CTR.500203
                    If Tax rate 2019-2020 is provided, then a valid range for Tax rate 2019-2020 is 0.0001 - 0.45000

                Legacy Rule Format:
                    IF [CTR409] <> NULL AND ([CTR409] <> SET(0.0001- 0.45000))
                       RETURN VALIDATION MESSAGE
                    ENDIF

            Technical Business Rule Format:
                    ^CTR409 <> NULL AND (NotInSet(^CTR409, '0.0001-0.45000'))

        Data Elements:
    
            ^CTR409 = RP.Y0-1:TaxRate.Designation.Rate
            */
        protected void VRATOCTR500203(List<ProcessMessageDocument> response, CTR2022 report)
        {
            ProcessMessageDocument processMessage;
            bool assertion;

            assertion = (report.CTR409 != null && (report.CTR409.GetValueOrDefault() < 0.0001M || report.CTR409.GetValueOrDefault() > 0.45000M));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.CTR.500203",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Valid range for Tax rate 2019-2020 is 0.0001 - 0.45000",
                    Location = "/xbrli:xbrl/tns:TaxRate.Designation.Rate",//[@contextRef='" + report.RPY0Minus1Id + "']",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.CTR.500203" } },
                };

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "CTR409", Value = GetValueOrEmpty(report.CTR409) });

                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.CTR.500203

        #region VR.ATO.CTR.500204

        /*  VR.ATO.CTR.500204
                    If Tax rate 2020-2021 is provided, then a valid range for Tax rate 2020-2021 is 0.0001 - 0.45000

                Legacy Rule Format:
                    IF [CTR410] <> NULL AND ([CTR410] <> SET(0.0001,0.45000))
                       RETURN VALIDATION MESSAGE
                    ENDIF

            Technical Business Rule Format:
                    ^CTR410 <> NULL AND (NotInSet(^CTR410, '0.0001,0.45000'))

        Data Elements:
    
            ^CTR410 = RP:TaxRate.Designation.Rate
            */
        protected void VRATOCTR500204(List<ProcessMessageDocument> response, CTR2022 report)
        {
            ProcessMessageDocument processMessage;
            bool assertion;

            assertion = (report.CTR410 != null && (report.CTR410.GetValueOrDefault() < 0.0001M || report.CTR410.GetValueOrDefault() > 0.45000M));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.CTR.500204",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Valid range for Tax rate 2020-2021 is 0.0001 - 0.45000",
                    Location = "/xbrli:xbrl/tns:TaxRate.Designation.Rate",//[@contextRef='" + report.RPId + "']",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.CTR.500204" } },
                };

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "CTR410", Value = GetValueOrEmpty(report.CTR410) });



                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.CTR.500204


        #region VR.ATO.CTR.500317

        /*  VR.ATO.CTR.500317
                   If 2022 lodgment and Loss Carry Back Tax Offset completed:
                   it must be less than or equal to the least of :
                   - Closing franking account balance OR
                   - Total amount of Income Tax liability 2018-19, Income Tax liability 2019-20 and Income Tax liability 2020-21   OR
                   - 30% of total amount of Tax loss 2019-20 carried back to 2018-19, Tax loss 2020-21 carried back to 2018-19, Tax loss 2020-21 carried back to 2019-20, Tax loss 2021-22 carried back to 2018-19, Tax loss 2021-22 carried back to 2019-20 and Tax loss 2021-22 carried back to 2020-21

               Legacy Rule Format:
                   IF [CTR2] = 2022 AND [CTR413] <> NULL AND [CTR413] > MIN(([CTR139]),([CTR406]+[CTR407]+[CTR408]),(0.3*([CTR397]+[CTR398]+[CTR399]+[CTR400]+[CTR401]+[CTR402])))
                      RETURN VALIDATION MESSAGE
                   ENDIF


                   Note: MIN is a function that will return the least value of 2 or more values specified. The values specified can be a field value or a calculated value.

           Technical Business Rule Format:
                   ^CTR2 = 2022 AND ^CTR413 <> NULL AND ^CTR413 > MIN(^CTR139,(^CTR406+^CTR407+^CTR408),(0.3*(^CTR397+^CTR398+^CTR399+^CTR400+^CTR401+^CTR402)))            Note: MIN is a function that will return the least value of 2 or more values specified. The values specified can be a field value or a calculated value.

       Data Elements:

           ^CTR413 = RP:Tax.Losses.CarryBackOffset.Amount

           ^CTR2 = RP:Report.TargetFinancial.Year

           ^CTR139 = RP.Closing:Equity.Dividends.FrankingAccountBalance.Amount

           ^CTR406 = RP.Y0-3:IncomeTax.Liability.Amount

           ^CTR407 = RP.Y0-2:IncomeTax.Liability.Amount

           ^CTR408 = RP.Y0-1:IncomeTax.Liability.Amount

           ^CTR397 = RP.Y0Minus2toY0Minus3:Tax.Losses.Total.Amount

           ^CTR398 = RP.Y0Minus1toY0Minus3:Tax.Losses.Total.Amount

           ^CTR399 = RP.Y0Minus1toY0Minus2:Tax.Losses.Total.Amount

           ^CTR400 = RP.Y0toY0Minus3:Tax.Losses.Total.Amount

           ^CTR401 = RP.Y0toY0Minus2:Tax.Losses.Total.Amount

           ^CTR402 = RP.Y0toY0Minus1:Tax.Losses.Total.Amount
           */
        protected void VRATOCTR500317(List<ProcessMessageDocument> response, CTR2022 report)
        {
            ProcessMessageDocument processMessage;
            decimal minvalue;
            decimal Sum1 = 0.3M * (report.CTR397.GetValueOrDefault() + report.CTR398.GetValueOrDefault() + report.CTR399.GetValueOrDefault() + report.CTR400.GetValueOrDefault() + report.CTR401.GetValueOrDefault() + report.CTR402.GetValueOrDefault());
            decimal Sum2 = report.CTR406.GetValueOrDefault() + report.CTR407.GetValueOrDefault() + report.CTR408.GetValueOrDefault();

            //   minvalue = Math.Min(report.CTR139.GetValueOrDefault(), Math.Min(Sum1, Sum2));
            if (report.CTR397.HasValue || report.CTR398.HasValue || report.CTR399.HasValue || report.CTR400.HasValue || report.CTR401.HasValue || report.CTR402.HasValue)
            {
                minvalue = Sum1;
                if (report.CTR406.HasValue || report.CTR407.HasValue || report.CTR408.HasValue)
                {
                    minvalue = Math.Min(Sum1, Sum2);
                    if (report.CTR139.HasValue)
                        minvalue = Math.Min(report.CTR139.GetValueOrDefault(), minvalue);
                }
                else if (report.CTR139.HasValue)
                {
                    minvalue = Math.Min(Sum1, report.CTR139.GetValueOrDefault());

                }

            }
            else if (report.CTR406.HasValue || report.CTR407.HasValue || report.CTR408.HasValue)
            {
                minvalue = Sum2;
                if (report.CTR139.HasValue)
                    minvalue = Math.Min(report.CTR139.GetValueOrDefault(), minvalue);
            }
            else
            {
                minvalue = report.CTR139.GetValueOrDefault();
            }

            bool isNotAllMinElementNull = (report.CTR139.HasValue || report.CTR397.HasValue || report.CTR398.HasValue || report.CTR399.HasValue || report.CTR400.HasValue || report.CTR401.HasValue || report.CTR402.HasValue ||
                report.CTR406.HasValue || report.CTR407.HasValue || report.CTR408.HasValue);

            bool assertion = (report.CTR2.GetValueOrDefault() == 2022 && report.CTR413 != null && isNotAllMinElementNull && report.CTR413.GetValueOrDefault() > minvalue);

            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.CTR.500317",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"If Loss carry back tax offset is provided, it must be less than or equal to the least of Closing franking account balance, or total income tax liability in 18/19,19/20, 20/21 or 30% of total tax loss carried back amounts",
                    LongDescription = @"If Loss carry back tax offset is provided, it must be less than or equal to the least of Closing franking account balance, or total amount of Income Tax liability 2018-19, Income Tax liability 2019-20 and Income Tax liability 2020-21, or 30% of total amount of Tax loss 2019-20 carried back to 2018-19, Tax loss 2020-21 carried back to 2018-19, Tax loss 2020-21 carried back to 2019-20, Tax loss 2021-22 carried back to 2018-19, Tax loss 2021-22 carried back to 2019-20 and Tax loss 2021-22 carried back to 2020-21",
                    Location = "/xbrli:xbrl/tns:Tax.Losses.CarryBackOffset.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.CTR.500317" } },
                };


                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "CTR2", Value = GetValueOrEmpty(report.CTR2) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "CTR413", Value = GetValueOrEmpty(report.CTR413) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "CTR406", Value = GetValueOrEmpty(report.CTR406) });
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "CTR407", Value = GetValueOrEmpty(report.CTR407) });
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "CTR408", Value = GetValueOrEmpty(report.CTR408) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "CTR397", Value = GetValueOrEmpty(report.CTR397) });
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "CTR398", Value = GetValueOrEmpty(report.CTR398) });
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "CTR399", Value = GetValueOrEmpty(report.CTR399) });
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "CTR400", Value = GetValueOrEmpty(report.CTR400) });
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "CTR401", Value = GetValueOrEmpty(report.CTR401) });
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "CTR402", Value = GetValueOrEmpty(report.CTR402) });

                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.CTR.500317
        #endregion Manual Rules - Rules marked for manual coding
    }
}

