using System.Linq;
using System.Collections.Generic;
using Ato.CD.Inbound.Shared;

using VaTS;
using Ato.EN.IntegrationServices.CodeGenerationCTR;
using DataContracts;
using Ato.EN.IntegrationServices.CodeGenerationRDTIS;


namespace Ato.CD.Inbound.CTR202202
{
    internal class CrossFormValidatorRDTIS : ICrossFormValidator
    {
        private CTR2022 ParentDocument { get; }
        private RDTIS2022 ChildDocument { get; }

        internal CrossFormValidatorRDTIS(CTR2022 report, BusinessDocument childDocument)
        {          
            ParentDocument = report;
            RDTIS2022XmlConsumer consumer = new RDTIS2022XmlConsumer();
            childDocument.DocumentStream.Position = 0;
            ChildDocument = consumer.Consume(childDocument.DocumentStream, true); //childDocument.ConsumedReport;
        }
           

        public IEnumerable<ProcessMessageDocument> ValidateCrossFormRules()
        {
            var response = new List<ProcessMessageDocument>();
           
            VRATOGEN402009(response);
            VRATOIEE438104(response);

            return response;
        }


        //RTDIS_CTR
        #region VR.ATO.GEN.402009
        /*
         *            IF (RP:entity.identifier.TFN <> PARENT RETURN:RP:entity.identifier.TFN)
            RETURN VALIDATION MESSAGE
            ENDIF
      */

        private void VRATOGEN402009(List<ProcessMessageDocument> response)
        {
            bool assertion = !ParentDocument.RPIdentifierTFN.Equals(ChildDocument.RDTIS70);
           

            if (assertion)
            {
                var processMessage = new ProcessMessageDocument
                {
                    Code = "CMN.ATO.GEN.402009",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Your supplied TFN does not match the TFN supplied on the form it was submitted with",
                    LongDescription = @"Your supplied TFN does not match the TFN supplied on the parent form it was submitted with",
                    Location = "/tns:RDTIS/tns:RP/tns:TaxFileNumberId",
                    Parameters = new ProcessMessageParameters { new ProcessMessageParameter { Name = "RuleIdentifier", Value = "VR.ATO.GEN.402009" } }
                };
                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.GEN.402009

        #region  VR.ATO.RDTIS.438104
        /*  
             VR.ATO.RDTIS.438104
        (^RDTIS71 <> PARENT PERIOD START DATE) OR (^RDTIS72 <> PARENT PERIOD END DATE)

         */
        protected void VRATOIEE438104(List<ProcessMessageDocument> response)
        {
            ProcessMessageDocument processMessage;
            bool assertion = ((ChildDocument.RDTIS71.HasValue && ChildDocument.RDTIS71 != ParentDocument.RPStartDate) || ChildDocument.RDTIS72.HasValue && ChildDocument.RDTIS72 != ParentDocument.RPEndDate);

            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.RDTIS.438104",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Reporting period does not match the reporting period supplied on the form it was submitted with",
                    Location = "/tns:RDTIS/tns:RP/tns:PeriodStartD",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.RDTIS.438104" } },
                };

                processMessage.Parameters.Add(new ProcessMessageParameter { Name = "RDTIS71", Value = ChildDocument.RDTIS71.ToString() });
                processMessage.Parameters.Add(new ProcessMessageParameter { Name = "RDTIS72", Value = ChildDocument.RDTIS72.ToString() });
                response.Add(processMessage);
            }
        }

        #endregion // VR.ATO.GEN.438000

   
    }
}