﻿using System.Linq;
using System.Collections.Generic;
using VaTS;
using DataContracts;
using Ato.EN.IntegrationServices.CodeGenerationTRTAMI; 
using Ato.EN.IntegrationServices.CodeGenerationCGTS;

namespace Ato.CD.Inbound.TRTAMI202202
{
    internal class CrossFormValidator 
    {
        private TRTAMI2022 ParentDocument { get; }
        private BusinessDocumentCollection ChildDocuments { get; }

        public CrossFormValidator(TRTAMI2022 report, BusinessDocumentCollection childDocuments)
        {
            ParentDocument = report;
            ChildDocuments = childDocuments ?? new BusinessDocumentCollection();
        }

        public List<ProcessMessageDocument> ValidateCrossFormRules()
        {
            var response = new List<ProcessMessageDocument>();
             
            VRATOTRTAMI000235(response);
            VRATOTRTAMI000203(response); 
            return response;
        }

 

        #region VR.ATO.TRTAMI.000235
        /*  VR.ATO.TRTAMI.000235
            IF COUNT(SCHEDULE = "CGTS") > 1 AND [CGTS153] = ANY OTHER OCCURRENCE OF [CGTS153]
               RETURN VALIDATION MESSAGE
            ENDIF
        */
        private void VRATOTRTAMI000235(List<ProcessMessageDocument> response)
        {
            IEnumerable<CGTS2018> duplicates = ChildDocuments.Where(s => s.DocumentName.Equals(Schedule.CGTS.ToString()))
                .Select(c => (CGTS2018) c.ConsumedReport).GroupBy(s => s.CGTS153).Where(s => s.Skip(1).Any())
                .SelectMany(c => c);

            response.AddRange(from duplicate in duplicates
                              select new ProcessMessageDocument
                              {
                                  Code = "CMN.ATO.TRTAMI.000235",
                                  Severity = ProcessMessageSeverity.Error,
                                  Description = "Name of AMIT Class for CGT schedule is invalid",
                                  LongDescription = "When multiple Capital Gains Tax (CGT) schedules are provided, a unique Name of AMIT Class must be provided for each",
                                  Location = "/xbrli:xbrl/xbrli:context",
                                  Parameters = new ProcessMessageParameters { new ProcessMessageParameter { Name = "RuleIdentifier", Value = "VR.ATO.TRTAMI.000235" } }
                              });
        }
        #endregion // VR.ATO.TRTAMI.000235

        #region  VR.ATO.TRTAMI.000203
        /*
           Rule ID: VR.ATO.TRTAMI.000203
           IF COUNT(SCHEDULE = "PSS") > 1
             RETURN VALIDATION MESSAGE
           ENDIF
       */
       private void VRATOTRTAMI000203(List<ProcessMessageDocument> response)
        {
            bool assertion = (ChildDocuments.Count(c => c.DocumentName.Equals("PSS")) > 1);

            if (assertion)
            {
                var processMessage = new ProcessMessageDocument
                {
                    Code = "CMN.ATO.TRTAMI.000203",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"The number of PAYG Payment Summary Schedules must not exceed 1",
                    Location = "/xbrli:xbrl/xbrli:context",
                    Parameters = new ProcessMessageParameters { new ProcessMessageParameter { Name = "RuleIdentifier", Value = "CMN.ATO.TRTAMI.000203" } }
                };
                response.Add(processMessage);
            }
        }
        #endregion
  

    }
}