using System.Linq;
using System.Collections.Generic;
using VaTS;
using DataContracts;
using System;
using Ato.EN.IntegrationServices.CodeGenerationIDS;
using Ato.EN.IntegrationServices.CodeGenerationTRTAMI;
using System.Text.RegularExpressions;

namespace Ato.CD.Inbound.TRTAMI202202
{
    internal class CrossFormValidatorIDS : ICrossFormValidator
    {
        private TRTAMI2022 ParentDocument { get; }
        private IDS2022 ChildDocument { get; }
        private List<Context> Contexts { get; }

        internal CrossFormValidatorIDS(TRTAMI2022 report, BusinessDocument childDocument)
        {
            ParentDocument = report;
            ChildDocument = (IDS2022)childDocument.ConsumedReport;
        }

        public List<ProcessMessageDocument> ValidateCrossFormRules()
        {
            var response = new List<ProcessMessageDocument>();

            VRATOIDS440670(response);
            return response;
        }


        #region VR.ATO.IDS.440670

        /*  VR.ATO.IDS.440670
                    Your supplied TFN does not match the TFN supplied on the form it was submitted with

                Legacy Rule Format:
                    IF (RP:entity.identifier.TFN <> PARENT RETURN:RP:entity.identifier.TFN)
                    RETURN VALIDATION MESSAGE
                    ENDIF

            Technical Business Rule Format:
                    ^TFN <> ^ParentTFN

        Data Elements:
            
                    ^ParentTFN = ParentTFN
            
                    ^TFN = RP
                    */
        private void VRATOIDS440670(List<ProcessMessageDocument> response)
        {
            bool assertion = !ChildDocument.IDS500.Equals(ParentDocument.RPIdentifierTFN);

            if (assertion)
            {
                var processMessage = new ProcessMessageDocument
                {
                    Code = "CMN.ATO.GEN.402009",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Your supplied TFN does not match the TFN supplied on the form it was submitted with",
                    LongDescription = @"Your supplied TFN does not match the TFN supplied on the parent form it was submitted with",
                    Location = "/xbrli:xbrl/xbrli:context/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters { new ProcessMessageParameter { Name = "RuleIdentifier", Value = "VR.ATO.IDS.440670" } }
                };

                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.IDS.440670

    }
}
